@@ -9,15 +9,18 @@
     :name: Microsoft SQL Server
 
 
-Auto Increment Behavior
------------------------
+.. _mssql_identity:
+
+Auto Increment Behavior / IDENTITY Columns
+------------------------------------------
 
 SQL Server provides so-called "auto incrementing" behavior using the
-``IDENTITY`` construct, which can be placed on an integer primary key.
-SQLAlchemy considers ``IDENTITY`` within its default "autoincrement" behavior,
-described at :paramref:`.Column.autoincrement`; this means
-that by default, the first integer primary key column in a :class:`.Table`
-will be considered to be the identity column and will generate DDL as such::
+``IDENTITY`` construct, which can be placed on any single integer column in a
+table. SQLAlchemy considers ``IDENTITY`` within its default "autoincrement"
+behavior for an integer primary key column, described at
+:paramref:`.Column.autoincrement`.  This means that by default, the first
+integer primary key column in a :class:`.Table` will be considered to be the
+identity column and will generate DDL as such::
 
     from sqlalchemy import Table, MetaData, Column, Integer
 
@@ -38,7 +41,8 @@ The above example will generate DDL as:
     )
 
 For the case where this default generation of ``IDENTITY`` is not desired,
-specify ``autoincrement=False`` on all integer primary key columns::
+specify ``False`` for the :paramref:`.Column.autoincrement` flag,
+on the first integer primary key column::
 
     m = MetaData()
     t = Table('t', m,
@@ -46,34 +50,69 @@ specify ``autoincrement=False`` on all integer primary key columns::
             Column('x', Integer))
     m.create_all(engine)
 
+To add the ``IDENTITY`` keyword to a non-primary key column, specify
+``True`` for the :paramref:`.Column.autoincrement` flag on the desired
+:class:`.Column` object, and ensure that :paramref:`.Column.autoincrement`
+is set to ``False`` on any integer primary key column::
+
+    m = MetaData()
+    t = Table('t', m,
+            Column('id', Integer, primary_key=True, autoincrement=False),
+            Column('x', Integer, autoincrement=True))
+    m.create_all(engine)
+
+.. versionchanged::  1.3   Added ``mssql_identity_start`` and
+   ``mssql_identity_increment`` parameters to :class:`.Column`.  These replace
+   the use of the :class:`.Sequence` object in order to specify these values.
+
+.. deprecated:: 1.3
+
+   The use of :class:`.Sequence` to specify IDENTITY characteristics is
+   deprecated and will be removed in a future release.   Please use
+   the ``mssql_identity_start`` and ``mssql_identity_increment`` parameters
+   documented at :ref:`mssql_identity`.
+
 .. note::
 
-    An INSERT statement which refers to an explicit value for such
-    a column is prohibited by SQL Server, however SQLAlchemy will detect this
-    and modify the ``IDENTITY_INSERT`` flag accordingly at statement execution
-    time.  As this is not a high performing process, care should be taken to
-    set the ``autoincrement`` flag appropriately for columns that will not
-    actually require IDENTITY behavior.
+    There can only be one IDENTITY column on the table.  When using
+    ``autoincrement=True`` to enable the IDENTITY keyword, SQLAlchemy does not
+    guard against multiple columns specifying the option simultaneously.  The
+    SQL Server database will instead reject the ``CREATE TABLE`` statement.
+
+.. note::
+
+    An INSERT statement which attempts to provide a value for a column that is
+    marked with IDENTITY will be rejected by SQL Server.   In order for the
+    value to be accepted, a session-level option "SET IDENTITY_INSERT" must be
+    enabled.   The SQLAlchemy SQL Server dialect will perform this operation
+    automatically when using a core :class:`.Insert` construct; if the
+    execution specifies a value for the IDENTITY column, the "IDENTITY_INSERT"
+    option will be enabled for the span of that statement's invocation.However,
+    this scenario is not high performing and should not be relied upon for
+    normal use.   If a table doesn't actually require IDENTITY behavior in its
+    integer primary key column, the keyword should be disabled when creating
+    the table by ensuring that ``autoincrement=False`` is set.
 
 Controlling "Start" and "Increment"
 ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 
-Specific control over the parameters of the ``IDENTITY`` value is supported
-using the :class:`.schema.Sequence` object.  While this object normally
-represents an explicit "sequence" for supporting backends, on SQL Server it is
-re-purposed to specify behavior regarding the identity column, including
-support of the "start" and "increment" values::
+Specific control over the "start" and "increment" values for
+the ``IDENTITY`` generator are provided using the
+``mssql_identity_start`` and ``mssql_identity_increment`` parameters
+passed to the :class:`.Column` object::
 
-    from sqlalchemy import Table, Integer, Sequence, Column
+    from sqlalchemy import Table, Integer, Column
 
-    Table('test', metadata,
-           Column('id', Integer,
-                  Sequence('blah', start=100, increment=10),
-                  primary_key=True),
-           Column('name', String(20))
-         ).create(some_engine)
+    test = Table(
+        'test', metadata,
+        Column(
+            'id', Integer, primary_key=True, mssql_identity_start=100,
+             mssql_identity_increment=10
+        ),
+        Column('name', String(20))
+    )
 
-would yield:
+The CREATE TABLE for the above :class:`.Table` object would be:
 
 .. sourcecode:: sql
 
@@ -82,8 +121,12 @@ would yield:
      name VARCHAR(20) NULL,
      )
 
-Note that the ``start`` and ``increment`` values for sequences are
-optional and will default to 1,1.
+.. versionchanged:: 1.3  The ``mssql_identity_start`` and
+   ``mssql_identity_increment`` parameters are now used to affect the
+   ``IDENTITY`` generator for a :class:`.Column` under  SQL Server.
+   Previously, the :class:`.Sequence` object was used.  As SQL Server now
+   supports real sequences as a separate construct, :class:`.Sequence` will be
+   functional in the normal way in a future SQLAlchemy version.
 
 INSERT behavior
 ^^^^^^^^^^^^^^^^
@@ -199,9 +242,6 @@ CREATE TABLE statement for this column will yield::
 
     login VARCHAR(32) COLLATE Latin1_General_CI_AS NULL
 
-.. versionadded:: 0.8 Character collations are now part of the base string
-   types.
-
 LIMIT/OFFSET Support
 --------------------
 
@@ -500,8 +540,6 @@ names::
 
 would render the index as ``CREATE INDEX my_index ON table (x) INCLUDE (y)``
 
-.. versionadded:: 0.8
-
 Index ordering
 ^^^^^^^^^^^^^^
 
@@ -511,8 +549,6 @@ Index ordering is available via functional expressions, such as::
 
 would render the index as ``CREATE INDEX my_index ON table (x DESC)``
 
-.. versionadded:: 0.8
-
 .. seealso::
 
     :ref:`schema_indexes_functional`
@@ -525,7 +561,7 @@ is compatible with SQL2000 while running on a SQL2005 database
 server. ``server_version_info`` will always return the database
 server version information (in this case SQL2005) and not the
 compatibility level information. Because of this, if running under
-a backwards compatibility mode SQAlchemy may attempt to use T-SQL
+a backwards compatibility mode SQLAlchemy may attempt to use T-SQL
 statements that are unable to be parsed by the database server.
 
 Triggers
@@ -641,6 +677,7 @@ from ...types import TEXT
 from ...types import VARCHAR
 from ...util import update_wrapper
 
+
 # http://sqlserverbuilds.blogspot.com/
 MS_2016_VERSION = (13,)
 MS_2014_VERSION = (12,)
@@ -959,12 +996,26 @@ class DATETIMEOFFSET(sqltypes.TypeEngine):
         self.precision = precision
 
 
-class _StringType(object):
+class _UnicodeLiteral(object):
+    def literal_processor(self, dialect):
+        def process(value):
 
-    """Base for MSSQL string types."""
+            value = value.replace("'", "''")
 
-    def __init__(self, collation=None):
-        super(_StringType, self).__init__(collation=collation)
+            if dialect.identifier_preparer._double_percents:
+                value = value.replace("%", "%%")
+
+            return "N'%s'" % value
+
+        return process
+
+
+class _MSUnicode(_UnicodeLiteral, sqltypes.Unicode):
+    pass
+
+
+class _MSUnicodeText(_UnicodeLiteral, sqltypes.UnicodeText):
+    pass
 
 
 class TIMESTAMP(sqltypes._Binary):
@@ -1329,13 +1380,21 @@ class MSExecutionContext(default.DefaultExecutionContext):
                     and (
                         (
                             self.compiled.statement._has_multi_parameters
-                            and seq_column.key
-                            in self.compiled.statement.parameters[0]
+                            and (
+                                seq_column.key
+                                in self.compiled.statement.parameters[0]
+                                or seq_column
+                                in self.compiled.statement.parameters[0]
+                            )
                         )
                         or (
                             not self.compiled.statement._has_multi_parameters
-                            and seq_column.key
-                            in self.compiled.statement.parameters
+                            and (
+                                seq_column.key
+                                in self.compiled.statement.parameters
+                                or seq_column
+                                in self.compiled.statement.parameters
+                            )
                         )
                     )
                 )
@@ -1737,6 +1796,9 @@ class MSSQLCompiler(compiler.SQLCompiler):
             for t in [from_table] + extra_froms
         )
 
+    def visit_empty_set_expr(self, type_):
+        return "SELECT 1 WHERE 1!=1"
+
 
 class MSSQLStrictCompiler(MSSQLCompiler):
 
@@ -1800,6 +1862,7 @@ class MSDDLCompiler(compiler.DDLCompiler):
                 not column.nullable
                 or column.primary_key
                 or isinstance(column.default, sa_schema.Sequence)
+                or column.autoincrement is True
             ):
                 colspec += " NOT NULL"
             else:
@@ -1814,6 +1877,21 @@ class MSDDLCompiler(compiler.DDLCompiler):
         # install an IDENTITY Sequence if we either a sequence or an implicit
         # IDENTITY column
         if isinstance(column.default, sa_schema.Sequence):
+
+            if (
+                column.default.start is not None
+                or column.default.increment is not None
+                or column is not column.table._autoincrement_column
+            ):
+                util.warn_deprecated(
+                    "Use of Sequence with SQL Server in order to affect the "
+                    "parameters of the IDENTITY value is deprecated, as "
+                    "Sequence "
+                    "will correspond to an actual SQL Server "
+                    "CREATE SEQUENCE in "
+                    "a future release.  Please use the mssql_identity_start "
+                    "and mssql_identity_increment parameters."
+                )
             if column.default.start == 0:
                 start = 0
             else:
@@ -1823,8 +1901,13 @@ class MSDDLCompiler(compiler.DDLCompiler):
                 start,
                 column.default.increment or 1,
             )
-        elif column is column.table._autoincrement_column:
-            colspec += " IDENTITY(1,1)"
+        elif (
+            column is column.table._autoincrement_column
+            or column.autoincrement is True
+        ):
+            start = column.dialect_options["mssql"]["identity_start"]
+            increment = column.dialect_options["mssql"]["identity_increment"]
+            colspec += " IDENTITY(%s,%s)" % (start, increment)
         else:
             default = self.get_column_default_string(column)
             if default is not None:
@@ -1944,14 +2027,24 @@ class MSIdentifierPreparer(compiler.IdentifierPreparer):
     def quote_schema(self, schema, force=None):
         """Prepare a quoted table and schema name."""
 
+        # need to re-implement the deprecation warning entirely
+        if force is not None:
+            # not using the util.deprecated_params() decorator in this
+            # case because of the additional function call overhead on this
+            # very performance-critical spot.
+            util.warn_deprecated(
+                "The IdentifierPreparer.quote_schema.force parameter is "
+                "deprecated and will be removed in a future release.  This "
+                "flag has no effect on the behavior of the "
+                "IdentifierPreparer.quote method; please refer to "
+                "quoted_name()."
+            )
+
         dbname, owner = _schema_elements(schema)
         if dbname:
-            result = "%s.%s" % (
-                self.quote(dbname, force),
-                self.quote(owner, force),
-            )
+            result = "%s.%s" % (self.quote(dbname), self.quote(owner))
         elif owner:
-            result = self.quote(owner, force)
+            result = self.quote(owner)
         else:
             result = ""
         return result
@@ -2056,6 +2149,8 @@ class MSDialect(default.DefaultDialect):
         sqltypes.DateTime: _MSDateTime,
         sqltypes.Date: _MSDate,
         sqltypes.Time: TIME,
+        sqltypes.Unicode: _MSUnicode,
+        sqltypes.UnicodeText: _MSUnicodeText,
     }
 
     engine_config_types = default.DefaultDialect.engine_config_types.union(
@@ -2080,6 +2175,7 @@ class MSDialect(default.DefaultDialect):
         (sa_schema.PrimaryKeyConstraint, {"clustered": None}),
         (sa_schema.UniqueConstraint, {"clustered": None}),
         (sa_schema.Index, {"clustered": None, "include": None}),
+        (sa_schema.Column, {"identity_start": 1, "identity_increment": 1}),
     ]
 
     def __init__(
@@ -2296,19 +2392,13 @@ class MSDialect(default.DefaultDialect):
                 "join sys.schemas as sch on sch.schema_id=tab.schema_id "
                 "where tab.name = :tabname "
                 "and sch.name=:schname "
-                "and ind.is_primary_key=0 and ind.type != 0",
-                bindparams=[
-                    sql.bindparam(
-                        "tabname",
-                        tablename,
-                        sqltypes.String(convert_unicode=True),
-                    ),
-                    sql.bindparam(
-                        "schname", owner, sqltypes.String(convert_unicode=True)
-                    ),
-                ],
-                typemap={"name": sqltypes.Unicode()},
+                "and ind.is_primary_key=0 and ind.type != 0"
             )
+            .bindparams(
+                sql.bindparam("tabname", tablename, sqltypes.String()),
+                sql.bindparam("schname", owner, sqltypes.String()),
+            )
+            .columns(name=sqltypes.Unicode())
         )
         indexes = {}
         for row in rp:
@@ -2327,19 +2417,13 @@ class MSDialect(default.DefaultDialect):
                 "ind_col.object_id=tab.object_id) "
                 "join sys.schemas as sch on sch.schema_id=tab.schema_id "
                 "where tab.name=:tabname "
-                "and sch.name=:schname",
-                bindparams=[
-                    sql.bindparam(
-                        "tabname",
-                        tablename,
-                        sqltypes.String(convert_unicode=True),
-                    ),
-                    sql.bindparam(
-                        "schname", owner, sqltypes.String(convert_unicode=True)
-                    ),
-                ],
-                typemap={"name": sqltypes.Unicode()},
+                "and sch.name=:schname"
             )
+            .bindparams(
+                sql.bindparam("tabname", tablename, sqltypes.String()),
+                sql.bindparam("schname", owner, sqltypes.String()),
+            )
+            .columns(name=sqltypes.Unicode())
         )
         for row in rp:
             if row["index_id"] in indexes:
@@ -2360,17 +2444,10 @@ class MSDialect(default.DefaultDialect):
                 " where "
                 "mod.object_id=views.object_id and "
                 "views.schema_id=sch.schema_id and "
-                "views.name=:viewname and sch.name=:schname",
-                bindparams=[
-                    sql.bindparam(
-                        "viewname",
-                        viewname,
-                        sqltypes.String(convert_unicode=True),
-                    ),
-                    sql.bindparam(
-                        "schname", owner, sqltypes.String(convert_unicode=True)
-                    ),
-                ],
+                "views.name=:viewname and sch.name=:schname"
+            ).bindparams(
+                sql.bindparam("viewname", viewname, sqltypes.String()),
+                sql.bindparam("schname", owner, sqltypes.String()),
             )
         )
 
@@ -2480,9 +2557,10 @@ class MSDialect(default.DefaultDialect):
             if type_name.endswith("identity") and col_name in colmap:
                 ic = col_name
                 colmap[col_name]["autoincrement"] = True
-                colmap[col_name]["sequence"] = dict(
-                    name="%s_identity" % col_name
-                )
+                colmap[col_name]["dialect_options"] = {
+                    "mssql_identity_start": 1,
+                    "mssql_identity_increment": 1,
+                }
                 break
         cursor.close()
 
@@ -2495,8 +2573,11 @@ class MSDialect(default.DefaultDialect):
 
             row = cursor.first()
             if row is not None and row[0] is not None:
-                colmap[ic]["sequence"].update(
-                    {"start": int(row[0]), "increment": int(row[1])}
+                colmap[ic]["dialect_options"].update(
+                    {
+                        "mssql_identity_start": int(row[0]),
+                        "mssql_identity_increment": int(row[1]),
+                    }
                 )
         return cols
 
