@@ -5,10 +5,12 @@ import datetime
 
 import sqlalchemy as sa
 from sqlalchemy import Boolean
+from sqlalchemy import Enum
 from sqlalchemy import event
 from sqlalchemy import ForeignKey
 from sqlalchemy import func
 from sqlalchemy import Integer
+from sqlalchemy import literal
 from sqlalchemy import literal_column
 from sqlalchemy import select
 from sqlalchemy import String
@@ -27,6 +29,7 @@ from sqlalchemy.testing.assertsql import AllOf
 from sqlalchemy.testing.assertsql import CompiledSQL
 from sqlalchemy.testing.schema import Column
 from sqlalchemy.testing.schema import Table
+from sqlalchemy.util import OrderedDict
 from sqlalchemy.util import u
 from sqlalchemy.util import ue
 from test.orm import _fixtures
@@ -476,6 +479,8 @@ class ForeignPKTest(fixtures.MappedTest):
 
 
 class ClauseAttributesTest(fixtures.MappedTest):
+    __backend__ = True
+
     @classmethod
     def define_tables(cls, metadata):
         Table(
@@ -497,6 +502,19 @@ class ClauseAttributesTest(fixtures.MappedTest):
             Column("value", Boolean),
         )
 
+        Table(
+            "pk_t",
+            metadata,
+            Column(
+                "p_id",
+                Integer,
+                key="id",
+                autoincrement=True,
+                primary_key=True,
+            ),
+            Column("data", String(30)),
+        )
+
     @classmethod
     def setup_classes(cls):
         class User(cls.Comparable):
@@ -505,12 +523,17 @@ class ClauseAttributesTest(fixtures.MappedTest):
         class HasBoolean(cls.Comparable):
             pass
 
+        class PkDefault(cls.Comparable):
+            pass
+
     @classmethod
     def setup_mappers(cls):
         User, users_t = cls.classes.User, cls.tables.users_t
         HasBoolean, boolean_t = cls.classes.HasBoolean, cls.tables.boolean_t
+        PkDefault, pk_t = cls.classes.PkDefault, cls.tables.pk_t
         mapper(User, users_t)
         mapper(HasBoolean, boolean_t)
+        mapper(PkDefault, pk_t)
 
     def test_update(self):
         User = self.classes.User
@@ -566,6 +589,19 @@ class ClauseAttributesTest(fixtures.MappedTest):
 
         assert (u.counter == 5) is True
 
+    @testing.requires.sql_expressions_inserted_as_primary_key
+    def test_insert_pk_expression(self):
+        PkDefault = self.classes.PkDefault
+
+        pk = PkDefault(id=literal(5) + 10, data="some data")
+        session = Session()
+        session.add(pk)
+        session.flush()
+
+        eq_(pk.id, 15)
+        session.commit()
+        eq_(pk.id, 15)
+
     def test_update_special_comparator(self):
         HasBoolean = self.classes.HasBoolean
 
@@ -889,8 +925,7 @@ class ExtraPassiveDeletesTest(fixtures.MappedTest):
         mc.children[0].data = "some new data"
         assert_raises(sa.exc.DBAPIError, session.flush)
 
-    def test_extra_passive_obj_removed_o2m_still_nulls_out(self):
-        # see #3844, which we decided was not a bug
+    def test_extra_passive_obj_removed_o2m(self):
         myothertable, MyClass, MyOtherClass, mytable = (
             self.tables.myothertable,
             self.classes.MyClass,
@@ -909,15 +944,20 @@ class ExtraPassiveDeletesTest(fixtures.MappedTest):
 
         session = create_session()
         mc = MyClass()
-        moc = MyOtherClass()
-        mc.children.append(moc)
-        session.add_all([mc, moc])
+        moc1 = MyOtherClass()
+        moc2 = MyOtherClass()
+        mc.children.append(moc1)
+        mc.children.append(moc2)
+        session.add_all([mc, moc1, moc2])
         session.flush()
 
-        mc.children.remove(moc)
+        mc.children.remove(moc1)
+        mc.children.remove(moc2)
+        moc1.data = "foo"
         session.flush()
 
-        eq_(moc.parent_id, None)
+        eq_(moc1.parent_id, mc.id)
+        eq_(moc2.parent_id, mc.id)
 
     def test_dont_emit(self):
         myothertable, MyClass, MyOtherClass, mytable = (
@@ -2654,17 +2694,13 @@ class ManyToManyTest(_fixtures.FixtureTest):
 
         mapper(Keyword, keywords)
 
-        # note that we are breaking a rule here, making a second
-        # mapper(Keyword, keywords) the reorganization of mapper construction
-        # affected this, but was fixed again
-
         mapper(
             IKAssociation,
             item_keywords,
             primary_key=[item_keywords.c.item_id, item_keywords.c.keyword_id],
             properties=dict(
                 keyword=relationship(
-                    mapper(Keyword, keywords, non_primary=True),
+                    Keyword,
                     lazy="joined",
                     uselist=False,
                     # note here is a valid place where
@@ -3338,3 +3374,93 @@ class PartialNullPKTest(fixtures.MappedTest):
 
         t.col1 = "1"
         s.commit()
+
+
+class EnsurePKSortableTest(fixtures.MappedTest):
+    class SomeEnum(object):
+        # Implements PEP 435 in the minimal fashion needed by SQLAlchemy
+        __members__ = OrderedDict()
+
+        def __init__(self, name, value, alias=None):
+            self.name = name
+            self.value = value
+            self.__members__[name] = self
+            setattr(self.__class__, name, self)
+            if alias:
+                self.__members__[alias] = self
+                setattr(self.__class__, alias, self)
+
+    class MySortableEnum(SomeEnum):
+        __members__ = OrderedDict()
+
+        def __lt__(self, other):
+            return self.value < other.value
+
+    class MyNotSortableEnum(SomeEnum):
+        __members__ = OrderedDict()
+
+    one = MySortableEnum("one", 1)
+    two = MySortableEnum("two", 2)
+    three = MyNotSortableEnum("three", 3)
+    four = MyNotSortableEnum("four", 4)
+
+    @classmethod
+    def define_tables(cls, metadata):
+        Table(
+            "t1",
+            metadata,
+            Column("id", Enum(cls.MySortableEnum), primary_key=True),
+            Column("data", String(10)),
+        )
+
+        Table(
+            "t2",
+            metadata,
+            Column("id", Enum(cls.MyNotSortableEnum), primary_key=True),
+            Column("data", String(10)),
+        )
+
+    @classmethod
+    def setup_classes(cls):
+        class T1(cls.Basic):
+            pass
+
+        class T2(cls.Basic):
+            pass
+
+    @classmethod
+    def setup_mappers(cls):
+        mapper(cls.classes.T1, cls.tables.t1)
+        mapper(cls.classes.T2, cls.tables.t2)
+
+    def test_exception_persistent_flush_py3k(self):
+        s = Session()
+
+        a, b = self.classes.T2(id=self.three), self.classes.T2(id=self.four)
+        s.add_all([a, b])
+        s.commit()
+
+        a.data = "bar"
+        b.data = "foo"
+        if sa.util.py3k:
+            assert_raises_message(
+                sa.exc.InvalidRequestError,
+                r"Could not sort objects by primary key; primary key values "
+                r"must be sortable in Python \(was: '<' not supported between "
+                r"instances of 'MyNotSortableEnum' and 'MyNotSortableEnum'\)",
+                s.flush,
+            )
+        else:
+            s.flush()
+        s.close()
+
+    def test_persistent_flush_sortable(self):
+        s = Session()
+
+        a, b = self.classes.T1(id=self.one), self.classes.T1(id=self.two)
+        s.add_all([a, b])
+        s.commit()
+
+        a.data = "bar"
+        b.data = "foo"
+        s.commit()
